/* Distributed Checksum Clearinghouse
 *
 * parse a named checksum
 *
 * Copyright (c) 2003 by Rhyolite Software
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Rhyolite Software DCC 1.2.14-1.27 $Revision$
 */

#include "dcc_ck.h"
#include <arpa/inet.h>


#define DCC_MAX_PARSE_SUMS 1024
typedef struct {
    int	    num;
    DCC_SUM sums[DCC_MAX_PARSE_SUMS];
} IP_SUMS;


static int				/* -1=fatal 0=failed 1=ok 2=resolved */
dcc_host2ck(DCC_EMSG emsg, DCC_WF *wf,
	    IP_SUMS *ipsums, const char *ck,
	    const char *fnm, int lineno,
	    DCC_PARSE_CK_LOCK unlock)	/* unlock file if resolve hosts */
{
	const struct hostent *hp;
	const void *v;
	int error;
	struct in6_addr addr6;
	int nsums, bits;

	/* recognize xxx.xxx.xxx.xxx/y for IP address blocks
	 * as well as simple IP addresses */
	bits = dcc_str2cidr(emsg, &addr6, 0, ck, fnm, lineno);
	if (bits < 0)
		return 0;
	if (bits > 0) {
		nsums = 1 << (128-bits);
		if (nsums > DIM(ipsums->sums)) {
			dcc_pemsg(EX_NOHOST, emsg,
				  "CIDR block length in %s too large%s",
				  ck, fnm_lineno(fnm, lineno));
			return 0;
		}

		for (ipsums->num = 0; ipsums->num < nsums; ++ipsums->num) {
			dcc_ck_ipv6(ipsums->sums[ipsums->num], &addr6);
			NTOHL(addr6.s6_addr32[3]);
			++addr6.s6_addr32[3];
			HTONL(addr6.s6_addr32[3]);
		}
		return 1;
	}

	if (wf->per_user) {
		dcc_pemsg(EX_NOHOST, emsg,
			  "hostname checksum illegal%s",
			  fnm_lineno(fnm, lineno));
		return 0;
	}

	/* Don't keep the table write-locked while we try to
	 * resolve hostnames into IP addresses.
	 * This unlocks the DCC_WF as well as the file, and so must
	 * not be done to private whitelists. */
	if (unlock) {
		if (!unlock(emsg, wf))
			return -1;
		dcc_wf_unlock(wf);
	}

	memset(&addr6, 0, sizeof(addr6));
	dcc_host_lock();
	/* don't use SOCKS because the only IP addresses it makes
	 * send to white-list are inside the SOCKS firewall */
	hp = dcc_get_host(ck, 1, &error);
	if (!hp) {
		dcc_pemsg(EX_NOHOST, emsg,
			  "hostname \"%s\": %s%s",
			  ck, hstrerror(error), fnm_lineno(fnm, lineno));
		dcc_freehostent();
		if (unlock)
			dcc_wf_lock(wf);
		return 0;
	}

	if (unlock)
		dcc_wf_lock(wf);
	for (ipsums->num = 0;
	     ipsums->num <= DIM(ipsums->sums)
	     && 0 != (v = hp->h_addr_list[ipsums->num]);
	     ++ipsums->num) {
		if (hp->h_addrtype == AF_INET6) {
			dcc_ck_ipv6(ipsums->sums[ipsums->num],
				    (struct in6_addr *)v);
		} else {
			dcc_toipv6(&addr6, *(struct in_addr *)v);
			dcc_ck_ipv6(ipsums->sums[ipsums->num], &addr6);
		}
	}
	dcc_freehostent();
	return 2;
}



/* generate checksum value from the name of the checksum and a string */
int					/* 2=resolved, 1=ok, 0=bad, -1=fatal */
dcc_parse_ck(DCC_EMSG emsg,		/* failure message here */
	     DCC_WF *wf,
	     const char *fnm,		/* 0 or location of cksum */
	     int lineno,
	     const char *type_nm,	/* type of checksum */
	     const char *str,		/* ASCII string to generate checksum */
	     DCC_TGTS tgts,		/* # of targets */
	     DCC_PARSE_CK_FNC fnc,	/* do something with the checksum */
	     DCC_PARSE_CK_LOCK unlock)	/* unlock for IP */
{
	DCC_CK_TYPES type;
	IP_SUMS ipsums;
	const char *pstr;
	char *phdr, c, hdr_buf[80];
	int resolved, i, j;

	/* look for the name of the checksum
	 * to find the checksumming function */
	type = dcc_str2type(type_nm);
	/* compute the checksum */
	switch (type) {
	case DCC_CK_IP:
		resolved = dcc_host2ck(emsg, wf, &ipsums, str,
				       fnm, lineno, unlock);
		break;

	case DCC_CK_ENV_FROM:
	case DCC_CK_FROM:
	case DCC_CK_MESSAGE_ID:
	case DCC_CK_RECEIVED:
	case DCC_CK_ENV_TO:
		ipsums.num = 1;
		dcc_str2ck(ipsums.sums[0], 0, 0, str);
		resolved = 1;
		break;

	case DCC_CK_SUB:
		str += strspn(str, DCC_WHITESPACE);
		pstr = str;
		phdr = hdr_buf;
		for (;;) {
			c = *pstr++;
			if (c == '\0' || c == ':'
			    || DCC_IS_WHITE(c))
				break;
			c = DCC_TO_LOWER(c);
			*phdr++ = c;
			if (phdr >= &hdr_buf[sizeof(hdr_buf)]) {
				dcc_pemsg(EX_DATAERR, emsg,
					  " imposible substitute header name"
					  " in \"%s\"%s",
					  str, fnm_lineno(fnm, lineno));
				return 0;
			}
		}
		pstr += strspn(pstr, DCC_WHITESPACE);
		if (*pstr == '\0' || phdr == hdr_buf) {
			dcc_pemsg(EX_DATAERR, emsg,
				  " substitute header name absent in \"%s\"%s",
				  str, fnm_lineno(fnm, lineno));
			return 0;
		}
		ipsums.num = 1;
		dcc_str2ck(ipsums.sums[0], hdr_buf, phdr-hdr_buf, pstr);
		resolved = 1;
		break;

	default:
		dcc_pemsg(EX_DATAERR, emsg,
			  "unrecognized checksum type \"%s\"%s",
			  type_nm, fnm_lineno(fnm, lineno));
		return 0;
	}
	if (resolved <= 0)
		return resolved;

	/* apply the function to the checksum,
	 * or to the checksums if the value is a hostname with several
	 * IP addresses */
	for (i = 0; i < ipsums.num; ++i) {
		j = fnc(emsg, wf, fnm, lineno, type, ipsums.sums[i], tgts);
		if (j <= 0)
			return j;
	}

	return resolved;
}



/* generate checksum value from the name of the checksum and hex values */
int					/* 1=ok, 0=bad checksum, -1=fatal */
dcc_parse_hex_ck(DCC_EMSG emsg,		/* failure message here */
		 DCC_WF *wf,
		 const char *fnm,	/* 0 or location of cksum */
		 int lineno,
		 const char *type_nm,	/* type of checksum */
		 const char *str,	/* ASCII string to generate checksum */
		 DCC_TGTS tgts,		/* # of targets */
		 DCC_PARSE_CK_FNC fnc)	/* do something with the checksum */
{
	DCC_CK_TYPES type;
	union {
	    u_int32_t n[4];
	    DCC_SUM sum;
	} u;

	type = dcc_str2type(type_nm);
	if (type == DCC_CK_INVALID) {
		dcc_pemsg(EX_DATAERR, emsg,
			  "unrecognized checksum type \"%s\"%s",
			  type_nm, fnm_lineno(fnm, lineno));
		return 0;
	}

	/* compute the checksum */
	if (4 != sscanf(str, "%16x %16x %16x %16x\n",
			&u.n[0], &u.n[1], &u.n[2], &u.n[3])
	    && (type != DCC_CK_SUB
		|| 4 != sscanf(str, "%*s %16x %16x %16x %16x\n",
			       &u.n[0], &u.n[1], &u.n[2], &u.n[3]))) {
		dcc_pemsg(EX_DATAERR, emsg,
			  "unrecognized checksum values \"%s\"%s",
			  str, fnm_lineno(fnm, lineno));
		return 0;
	}
	HTONL(u.n[0]);
	HTONL(u.n[1]);
	HTONL(u.n[2]);
	HTONL(u.n[3]);

	/* apply the function to the checksum */
	return fnc(emsg, wf, fnm, lineno, type, u.sum, tgts);
}
